#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h> // toupper()
#include <dos.h>
#include <i86.h>
#include <direct.h>
#include "dir.h"
#include "ui.h"

static char oldDrive, pathBuf[PATH_MAX+1], oldPath[PATH_MAX+1], tmpPath[PATH_MAX+1];

const char *Dir_CurPath = NULL;
char Dir_DriveList[26]; // ABC... (not terminated)
uint8_t Dir_NumDrives = 0;
dir_t *Dir_List = NULL;
uint32_t Dir_NumFiles = 0;

bool Dir_IsDir(uint32_t fileEntry)
{
	dir_t *d;
	if (fileEntry >= Dir_NumFiles)
		return false;

	return Dir_List[fileEntry].isDir;
}

void Dir_Init(void)
{
	Dir_GetDrives();
	oldDrive = Dir_GetDrive();
	Dir_GetPath();
	strcpy(oldPath, Dir_CurPath);
}

void Dir_Reset(void)
{
	Dir_SetDrive(oldDrive);
	Dir_SetPath(oldPath);
}

char Dir_GetDrive(void)
{
	union REGS r;

	r.w.ax = 0x1900;
	int386(0x21, &r, &r);

	return 'A' + r.h.al;
}

void Dir_GetPath(void) // -> Dir_CurPath
{
	union REGS r;
	struct SREGS sr;

	segread(&sr);
	r.w.ax = 0x4700;
	r.h.dl = 0;
	r.x.esi = FP_OFF(tmpPath);
	sr.ds = FP_SEG(tmpPath);
	int386x(0x21, &r, &r, &sr);

	if ((r.w.cflag & INTR_CF) == 0)
	{
		sprintf(pathBuf, "%c:\\%s", Dir_GetDrive(), tmpPath);
		Dir_CurPath = pathBuf;
	}
}

bool Dir_SetPath(const char *path)
{
	union REGS r;
	struct SREGS sr;

	segread(&sr);
	r.w.ax = 0x3B00;
	r.x.edx = FP_OFF(path);
	sr.ds = FP_SEG(path);
	int386x(0x21, &r, &r, &sr);
	
	if (r.w.cflag & INTR_CF)
		return false;

	return true;
}

bool Dir_SetDrive(char letter)
{
	union REGS r;
	struct SREGS sr;

	uint8_t oldDrive = Dir_GetDrive();

	r.w.ax = 0x0E00;
	r.h.dl = toupper(letter) - 65;
	int386(0x21, &r, &r);

	if (r.w.cflag & INTR_CF)
		return false;

	// test if the drive change really worked
	segread(&sr);
	r.w.ax = 0x4700;
	r.h.dl = 0;
	r.x.esi = FP_OFF(tmpPath);
	sr.ds = FP_SEG(tmpPath);
	int386x(0x21, &r, &r, &sr);

	if ((r.w.cflag & INTR_CF) != 0)
	{
		// attempt to set back the old drive
		r.w.ax = 0x0E00;
		r.h.dl = toupper(oldDrive) - 65;
		int386(0x21, &r, &r);
	
		return false;
	}

	return true;
}

void Dir_GetDrives(void) // -> Dir_NumDrives & Dir_DriveList
{
	uint8_t oldDrive, numDrives, i, drive;
	union REGS r;

	// get current drive
	r.w.ax = 0x1900;
	int386(0x21, &r, &r);
	oldDrive = r.h.al;

	// set default drive
	r.w.ax = 0x0E00;
	r.h.dl = oldDrive;
	int386(0x21, &r, &r);
	numDrives = r.h.al;

	Dir_NumDrives = 0;
	for (i = 0; i < numDrives; i++)
	{
		r.w.ax = 0x0E00;
		r.h.dl = (uint8_t)i;
		int386(0x21, &r, &r);

		r.w.ax = 0x1900;
		int386(0x21, &r, &r);
		drive = r.h.al;

		if (drive == (uint8_t)i)
		{
			// some kind of hack to verify that the drive letter is valid
			r.w.ax = 0x440E;
			r.h.bl = 1 + drive;
			int386(0x21, &r, &r);

			if ((r.w.cflag & INTR_CF) || r.h.al == 0 || r.h.al-1 == drive)
			{
				Dir_DriveList[Dir_NumDrives] = 'A'+drive;
				Dir_NumDrives++;
			}
		}
	}

	// set back old drive
	r.w.ax = 0x0E00;
	r.h.dl = oldDrive;
	int386(0x21, &r, &r);
}

void Dir_Free(void)
{
	if (Dir_List != NULL)
	{
		free(Dir_List);
		Dir_List = NULL;
	}

	Dir_NumFiles = 0;
}

static bool GrowList(void)
{
	dir_t *p = (dir_t *)realloc(Dir_List, (Dir_NumFiles + 1) * sizeof (dir_t));
	if (p != NULL)
	{
		Dir_List = p;
		Dir_NumFiles++;
		return true;
	}

	return false;
}

static bool AddFileEntry(struct find_t *src)
{
	char txtBuf[32], *fileExt;
	bool isDir;
	int32_t nameLen;
	uint32_t filesize;
	dir_t *dst;

	isDir = !!(src->attrib & _A_SUBDIR);

	nameLen = strlen(src->name);
	if (nameLen > 12)
		nameLen = 12;
		
	// entry filtering
	if (isDir)
	{
		if (nameLen == 1 && src->name[0] == '.' && src->name[1] == '\0')
			return true; // don't include folder

		if (nameLen == 2 && src->name[0] == '.' && src->name[1] == '.' && src->name[2] == '\0')
			return true; // don't include parent folder
	}
	else 
	{
		if (nameLen < 5)
			return true; // don't include file, its name is too short

		fileExt = src->name + (nameLen-4);
		if (fileExt[0] != '.')
			return true; // don't include file
		fileExt++;

		if (stricmp("mod", fileExt) != 0 && stricmp("stk", fileExt) != 0 && stricmp("ust", fileExt) != 0 && stricmp("nst", fileExt) != 0)
			return true; // don't include file
	}

	if (!GrowList())
	{
		Dir_Free();
		return false;
	}

	dst = &Dir_List[Dir_NumFiles-1];

	dst->isDrive = false;
	dst->isDir = isDir;
	dst->date[10] = '\0';
	dst->name[12] = '\0';
	dst->filesize[7] = '\0';

	memset(dst->name, ' ', 12);

	if (isDir)
	{
		memset(dst->date, ' ', 10);

		filesize = src->size;
		if (filesize > 9999999)
			filesize = 9999999;

		memcpy(dst->filesize, "  (DIR)", 7);
	}
	else
	{
		const uint8_t day = src->wr_date & 31;
		const uint8_t month = (src->wr_date >> 5) & 15;
		const uint16_t year = 1980 + ((src->wr_date >> 9) & 0x7F);
		
		sprintf(dst->date, "%02d/%02d/%04d", day, month, year);

		filesize = src->size;
		if (filesize > 9999999)
			filesize = 9999999;

		sprintf(dst->filesize, "%7d", filesize);
	}

	if (src->name != NULL)
		memcpy(dst->name, src->name, nameLen);

	return true;
}

bool Dir_Read(void)
{
	int32_t i;
	dir_t *p;
	struct find_t f;

	Dir_Free();

	// add disk drives
	for (i = 0; i < Dir_NumDrives; i++)
	{
		if (!GrowList())
			goto Error;

		p = &Dir_List[Dir_NumFiles-1];
		p->isDrive = true;

		memset(p->date, ' ', 10); p->date[10] = '\0';

		memset(p->name, ' ', 12); p->name[12] = '\0';
		p->name[0] = '[';
		p->name[1] = Dir_DriveList[i];
		p->name[2] = ']';	

		memcpy(p->filesize, "  (DRV)", 7); p->filesize[7] = '\0';
	}

	if (_dos_findfirst("*.*", _A_SUBDIR, &f) == -1)
		goto Done; // no files in folder

	if (!AddFileEntry(&f))
		goto Error;

	while (!_dos_findnext(&f))
	{
		if (!AddFileEntry(&f))
			goto Error;
	}

Done:
	return true;

Error:
	Dir_Free();
	return false;
}
