#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <conio.h>
#include <math.h>
#include "hardware/mouse.h"
#include "hardware/video.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "ui.h"
#include "asm.h"
#include "sampler.h"
#include "audio/paula.h"
#include "audio/replayer.h"

#define CLAMP(x, low, high) (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

#define SAMPLE_AREA_X 3
#define SAMPLE_AREA_Y 138
#define SAMPLE_AREA_H 48
#define SAMPLE_AREA_W 314
#define SAMPLE_AREA_Y_PLANEOFFSET (SAMPLE_AREA_Y*(SCREEN_W/8))
#define SAMPLE_AREA_Y_CENTER (SAMPLE_AREA_Y+23)
#define SAMPLE_AREA_BYTES (SAMPLE_AREA_H * (SCREEN_W/8))

sampler_t sampler; // globalized

static void WavePlaneLine(int32_t X0, int32_t Y0, int32_t X1, int32_t Y1);

#define WAVEPOS_BITS 23 /* max for 128kB samples */
#define WAVEPOS_MASK ((1 << WAVEPOS_BITS)-1)

static uint8_t waveScaleTable[256], *wavePlaneBuffer;
static uint16_t dragStart, dragEnd;
static int32_t oldSamplingLineXPos;
static uint32_t smpPosDelta;

bool Sampler_Init(void)
{
	int32_t i;
	for (i = 0; i < 256; i++)
		waveScaleTable[i] = 23 - (((int8_t)i * SAMPLE_AREA_H) >> 8);

	// do alloc so that we get 32-bit alignment
	wavePlaneBuffer = (uint8_t *)malloc(SAMPLE_AREA_BYTES * sizeof (uint32_t));
	if (wavePlaneBuffer == NULL)
		return false;
	
	return true;
}

void Sampler_Free(void)
{
	if (wavePlaneBuffer != NULL)
	{
		free(wavePlaneBuffer);
		wavePlaneBuffer = NULL;
	}
}

static void Sampler_SetDragBar(void)
{
	// clear drag bar background
	memset((uint8_t *)VGA_FRAMEBUFFER + (190 * (SCREEN_W/8)), 0, 4*(SCREEN_W/8));
	
	if (sampler.samLength > 0 && sampler.samDisplay != sampler.samLength)
	{
		int32_t pos;
		const int32_t roundingBias = sampler.samLength >> 1;
	
		// update drag bar coordinates
		pos = 4 + (((sampler.samOffset * 311) + roundingBias) / sampler.samLength);
		dragStart = (uint16_t)CLAMP(pos, 4, 315);

		pos = 5 + ((((sampler.samDisplay + sampler.samOffset) * 311) + roundingBias) / sampler.samLength);
		dragEnd = (uint16_t)CLAMP(pos, 5, 316);

		if (dragStart > dragEnd-1)
			dragStart = dragEnd-1;

		// draw drag bar
		if (dragEnd-dragStart > 0)
			GFX_FillRect(dragStart, 190, dragEnd, 190+4);
	}
}

void Sampler_CalculateDrawDelta(void)
{
	if (sampler.samDisplay == 0)
	{
		smpPosDelta = 0;
	}
	else
	{
		const uint32_t samDisplay = sampler.samDisplay;
		__asm
		{
			push eax
			push ebx
			push edx
			mov ebx,SAMPLE_AREA_W
			xor edx,edx
			mov eax,samDisplay
			shld edx,eax,WAVEPOS_BITS
			shl eax,WAVEPOS_BITS
			div ebx
			mov smpPosDelta,eax		
			pop edx
			pop ebx
			pop eax
		}
	}
}

static void Sampler_DrawDisplayLength(void)
{
	char dispString[6+1];
	uint32_t value = sampler.samDisplay;

	Video_SelectWritePlane(2);

	__asm
	{
		push eax
		push ebx
		push edx
		push edi
		// -------------------------------
		lea edi,dispString
		mov eax,value
		mov ebx,10
		// -------------------------------
		// digit #6
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+5],dl
		cmp eax,10
		jl p6zero1
		// -------------------------------
		// digit #5
		xor edx,edx
		div bx
		add dl,'0'
		mov [edi+4],dl
		test eax,eax
		jz p6zero2
		// -------------------------------
		// digit #4
		xor edx,edx
		div bx
		add dl,'0'
		mov [edi+3],dl
		test eax,eax
		jz p6zero3
		// -------------------------------
		// digit #3
		xor edx,edx
		div bx
		add dl,'0'
		mov [edi+2],dl
		test eax,eax
		jz p6zero4
		// -------------------------------
		// digit #2
		xor edx,edx
		div bx
		add dl,'0'
		mov [edi+1],dl
		test eax,eax
		jz p6zero5
		// -------------------------------
		// digit #1
		xor edx,edx
		div bx
		add dl,'0'
		mov [edi],dl
		jmp p6end
		// -------------------------------		
p6zero1:	mov byte ptr [edi+4],' '
p6zero2:	mov byte ptr [edi+3],' '
p6zero3:	mov byte ptr [edi+2],' '
p6zero4:	mov byte ptr [edi+1],' '
p6zero5:	mov byte ptr [edi+0],' '
p6end:		mov byte ptr [edi+6],0
		pop edi
		pop edx
		pop ebx
		pop eax
	}	
	
	Font_Print(264, 198, dispString);
}

void Sampler_UpdateSampleInfo(void)
{
	ptSmp_t *s;

	if (!ui.samplerScreenShown)
		return;

	s = &samples[ui.currSample];
	
	sampler.samStart = s->data;
	sampler.samOffset = 0;
	sampler.samLength = s->length << 1;
	sampler.samDisplay = sampler.samLength;
	
	Sampler_CalculateDrawDelta();
	Sampler_WriteWaveform();
	Sampler_DrawDisplayLength();
	Sampler_SetDragBar();
	
	// LOOP
	if (s->loopStart+s->loopLength <= 1)
		Font_Print(288, 209, "OFF");
	else
		Font_Print(288, 209, " ON");
}

void Sampler_EraseOldSamplingLine(void)
{
	if (ui.samplerScreenShown)
	{
		if (oldSamplingLineXPos >= SAMPLE_AREA_X && oldSamplingLineXPos < SAMPLE_AREA_X+SAMPLE_AREA_W)
			GFX_InvertSamplerLine(oldSamplingLineXPos);
	}
	
	oldSamplingLineXPos = -1;
}

void Sampler_WriteSamplingLine(void)
{
	int8_t *smpStart, *smpEnd;
	int32_t i;
	uint32_t currSamplingPos;
	ptSmp_t *s;
	paula_t *p;

	if (!ui.samplerScreenShown || sampler.samDisplay <= 0)
		return;
	
	currSamplingPos = UINT32_MAX;

	s = &samples[ui.currSample];
	smpStart = s->data;
	smpEnd = s->data + (s->length << 1);
	
	p = paula;
	for (i = 0; i < PAULA_VOICES; i++, p++)
	{
		if (p->dmaActive && p->dataPtr >= smpStart && p->dataPtr < smpEnd)
			currSamplingPos = (uint32_t)(p->dataPtr - smpStart);
	}

	if (currSamplingPos != UINT32_MAX)
	{
		int32_t xPos = SAMPLE_AREA_X+(((currSamplingPos - sampler.samOffset) * SAMPLE_AREA_W) / sampler.samDisplay);
		if (xPos >= SAMPLE_AREA_X && xPos < SAMPLE_AREA_X+SAMPLE_AREA_W)
		{
			GFX_InvertSamplerLine(xPos);
			oldSamplingLineXPos = xPos;
		}
	}
}

void Sampler_WriteWaveform(void)
{
	Video_SelectReadPlane(2); // needed for line routine
	Video_SelectWritePlane(2);
	
	if (sampler.samDisplay == 0 || sampler.samStart == NULL)
	{
		// draw empty sample
		ASM_FastFill32((uint8_t *)VGA_FRAMEBUFFER + SAMPLE_AREA_Y_PLANEOFFSET, 0, SAMPLE_AREA_BYTES/4);
		GFX_DrawLine(3, SAMPLE_AREA_Y_CENTER, 3+SAMPLE_AREA_W-1, SAMPLE_AREA_Y_CENTER, (uint8_t *)VGA_FRAMEBUFFER);
	}
	else
	{
		uint8_t *smpPtr = (uint8_t *)sampler.samStart + sampler.samOffset;
		int32_t x1, y1, x2, y2, x;
		uint32_t smpPosFrac;

		// clear sample bitplane buffer
		ASM_FastFill32(wavePlaneBuffer, 0, SAMPLE_AREA_H * (SCREEN_W/8));
		
		// plot sample to bitplane buffer
		smpPosFrac = 0;
		y1 = waveScaleTable[*smpPtr];
		for (x = 3; x < SAMPLE_AREA_W+2; x++)
		{
			smpPosFrac += smpPosDelta;
			smpPtr += smpPosFrac >> WAVEPOS_BITS;
			smpPosFrac &= WAVEPOS_MASK;
			
			y2 = waveScaleTable[*smpPtr];
			
			//GFX_DrawLine(x, y1, x+1, y2, wavePlaneBuffer);
			GFX_DrawSamplerLine(x, y1, y2, wavePlaneBuffer);

			y1 = y2;
		}

		// copy sample bitplane buffer to VRAM
		ASM_FastCopy32(wavePlaneBuffer,
			(uint8_t *)VGA_FRAMEBUFFER + SAMPLE_AREA_Y_PLANEOFFSET,
			SAMPLE_AREA_BYTES/4);
	}
}

void Sampler_Toggle(void)
{
	ui.samplerScreenShown ^= 1;
	UI_DrawBottomScreen();
	
	if (ui.samplerScreenShown)
	{
		Sampler_UpdateSampleInfo();
		Font_Print(288, 220, "C-3"); // sample play note
		Font_Print(288, 231, "C-3"); // sample resample note
	}
}
